from .shapes import tetrahedron, hexahedron, axis_aligned_cube, hexahedron_4pts, octahedron, icosahedron, dodecahedron, cylinder, sphere_uv, sphere_fibonacci, icosphere, torus
from .rings import ring, flat_ring
from .flat import triangle, quad, unit_grid, unit_triangle
from .polylines import chain_of_vertices, vector_field
from .transformations import spherify_vertices, cylindrify_edges
from .dual import dual_mesh